@@ -328,7 +328,7 @@ defer:
 void iofd_handle_segmented_read(struct osmo_io_fd *iofd, struct msgb *msg, int rc)
 {
 	int res;
-	struct msgb *pending = NULL;
+	struct msgb *pending;
 
 	OSMO_ASSERT(iofd->mode == OSMO_IO_FD_MODE_READ_WRITE);
 
@@ -338,9 +338,18 @@ void iofd_handle_segmented_read(struct osmo_io_fd *iofd, struct msgb *msg, int r
 	}
 
 	do {
+		pending = NULL;
 		res = iofd_handle_segmentation(iofd, msg, &pending);
-		if (res != IOFD_SEG_ACT_DEFER || rc < 0)
+		if (res != IOFD_SEG_ACT_DEFER || rc < 0) {
 			iofd->io_ops.read_cb(iofd, rc, msg);
+			/* The user could unregister/close the iofd during read_cb() above.
+			 * Once that's done, it doesn't expect to receive any more events,
+			 * so discard it: */
+			if (!IOFD_FLAG_ISSET(iofd, IOFD_FLAG_FD_REGISTERED)) {
+				msgb_free(pending);
+				return;
+			}
+		}
 		if (res == IOFD_SEG_ACT_HANDLE_MORE)
 			msg = pending;
 	} while (res == IOFD_SEG_ACT_HANDLE_MORE);
@@ -637,7 +646,7 @@ static int check_mode_callback_compat(enum osmo_io_fd_mode mode, const struct os
  *  The created osmo_io_fd is initialized with some default settings:
  *  * msgb allocations size: OSMO_IO_DEFAULT_MSGB_SIZE (1024)
  *  * msgb headroom: OSMO_IO_DEFAULT_MSGB_HEADROOM (128)
- *  * tx_queue depth: 32
+ *  * tx_queue depth: 1024
  *
  *  Those values may be adjusted from their defaults by using osmo_iofd_set_alloc_info() and
  *  osmo_iofd_set_txqueue_max_length() on the osmo_io_fd.
@@ -696,7 +705,7 @@ struct osmo_io_fd *osmo_iofd_setup(const void *ctx, int fd, const char *name, en
 	iofd->msgb_alloc.size = OSMO_IO_DEFAULT_MSGB_SIZE;
 	iofd->msgb_alloc.headroom = OSMO_IO_DEFAULT_MSGB_HEADROOM;
 
-	iofd->tx_queue.max_length = 32;
+	iofd->tx_queue.max_length = 1024;
 	INIT_LLIST_HEAD(&iofd->tx_queue.msg_queue);
 
 	return iofd;
@@ -736,19 +745,29 @@ int osmo_iofd_register(struct osmo_io_fd *iofd, int fd)
 {
 	int rc = 0;
 
-	if (fd >= 0)
-		iofd->fd = fd;
-	else if (iofd->fd < 0) {
+	if (fd < 0 && iofd->fd < 0) {
 		/* this might happen if both osmo_iofd_setup() and osmo_iofd_register() are called with -1 */
 		LOGPIO(iofd, LOGL_ERROR, "Cannot register io_fd using invalid fd == %d\n", iofd->fd);
 		return -EBADF;
 	}
+	if (fd < 0)
+		fd = iofd->fd;
+	else if (iofd->fd < 0)
+		iofd->fd = fd;
+
+	if (IOFD_FLAG_ISSET(iofd, IOFD_FLAG_FD_REGISTERED)) {
+		/* If re-registering same fd, handle as NO-OP.
+		 * New FD should go through unregister() first. */
+		return iofd->fd == fd ? 0 : -ENOTSUP;
+	}
 
 	rc = osmo_iofd_ops.register_fd(iofd);
 	if (rc)
 		return rc;
 
 	IOFD_FLAG_UNSET(iofd, IOFD_FLAG_CLOSED);
+	IOFD_FLAG_SET(iofd, IOFD_FLAG_FD_REGISTERED);
+
 	if ((iofd->mode == OSMO_IO_FD_MODE_READ_WRITE && iofd->io_ops.read_cb) ||
 	    (iofd->mode == OSMO_IO_FD_MODE_RECVFROM_SENDTO && iofd->io_ops.recvfrom_cb) ||
 	    (iofd->mode == OSMO_IO_FD_MODE_RECVMSG_SENDMSG && iofd->io_ops.recvmsg_cb)) {
@@ -771,7 +790,14 @@ int osmo_iofd_register(struct osmo_io_fd *iofd, int fd)
  */
 int osmo_iofd_unregister(struct osmo_io_fd *iofd)
 {
-	return osmo_iofd_ops.unregister_fd(iofd);
+	int rc;
+
+	if (!IOFD_FLAG_ISSET(iofd, IOFD_FLAG_FD_REGISTERED))
+		return 0;
+
+	rc = osmo_iofd_ops.unregister_fd(iofd);
+	IOFD_FLAG_UNSET(iofd, IOFD_FLAG_FD_REGISTERED);
+	return rc;
 }
 
 /*! Retrieve the number of messages pending in the transmit queue.
@@ -844,6 +870,10 @@ int osmo_iofd_close(struct osmo_io_fd *iofd)
 
 	iofd->pending = NULL;
 
+	osmo_iofd_ops.read_disable(iofd);
+	osmo_iofd_ops.write_disable(iofd);
+	osmo_iofd_unregister(iofd);
+
 	rc = osmo_iofd_ops.close(iofd);
 	iofd->fd = -1;
 	return rc;
@@ -994,12 +1024,22 @@ void osmo_iofd_get_ioops(struct osmo_io_fd *iofd, struct osmo_io_ops *ioops)
 
 /*! Request notification of the user if/when a client socket is connected.
  *  Calling this function will request osmo_io to notify the user (via
- *  write call-back) once a non-blocking outbound connect() of the
- *  socket completes.
+ *  write call-back with res=0 and msgb=NULL) once a non-blocking outbound
+ *  connect() of the socket completes.
  *
  *  This only works for connection oriented sockets in either
  *  OSMO_IO_FD_MODE_READ_WRITE or OSMO_IO_FD_MODE_RECVMSG_SENDMSG mode.
  *
+ * The fact that the write call-back is called with msgb=NULL can be used to
+ * distinguish before this "connected" notification and a socket write failure.
+ *
+ * If the server transmits data quick enough after accepting the connection,
+ * it may happen that a read call-back is triggered towards the user before this
+ * special write-callback, since both events may come together from the kernel.
+ * Hence under those scenarios where server starts the communication, it is
+ * important not to assume or require that the write-callback(res=0, msgb=NULL)
+ * will be the first one triggered.
+ *
  *  \param[in] iofd the file descriptor */
 void osmo_iofd_notify_connected(struct osmo_io_fd *iofd)
 {
